@@ -11,7 +11,7 @@ use std::{io, mem};
 use crate::fuse;
 
 use super::fs_cache_req_handler::FsCacheReqHandler;
-pub use fuse::{FsOptions, OpenOptions, RemovemappingOne, SetattrValid, ROOT_ID};
+pub use fuse::{FsOptions, OpenOptions, RemovemappingOne, SetattrValid, SetxattrFlags, ROOT_ID};
 
 /// Information about a path in the filesystem.
 pub struct Entry {
@@ -174,7 +174,7 @@ pub trait ZeroCopyReader {
     fn copy_to_end(&mut self, f: &mut File, mut off: u64) -> io::Result<usize> {
         let mut out = 0;
         loop {
-            match self.read_to(f, ::std::usize::MAX, off) {
+            match self.read_to(f, usize::MAX, off) {
                 Ok(0) => return Ok(out),
                 Ok(n) => {
                     off = off.saturating_add(n as u64);
@@ -267,7 +267,7 @@ pub trait ZeroCopyWriter {
     fn copy_to_end(&mut self, f: &mut File, mut off: u64) -> io::Result<usize> {
         let mut out = 0;
         loop {
-            match self.write_from(f, ::std::usize::MAX, off) {
+            match self.write_from(f, usize::MAX, off) {
                 Ok(0) => return Ok(out),
                 Ok(n) => {
                     off = off.saturating_add(n as u64);
@@ -432,7 +432,7 @@ pub trait FileSystem {
     /// The `valid` parameter indicates the fields of `attr` that may be considered valid and should
     /// be set by the file system. The content of all other fields in `attr` is undefined.
     ///
-    /// If the `FsOptions::HANDLE_KILLPRIV` was set during `init`, then the implementation is
+    /// If the `FsOptions::HANDLE_KILLPRIV_V2` was set during `init`, then the implementation is
     /// expected to reset the setuid and setgid bits if the file size or owner is being changed.
     ///
     /// This method returns the new attributes after making the modifications requested by the
@@ -863,6 +863,7 @@ pub trait FileSystem {
         name: &CStr,
         value: &[u8],
         flags: u32,
+        extra_flags: SetxattrFlags,
     ) -> io::Result<()> {
         Err(io::Error::from_raw_os_error(libc::ENOSYS))
     }
@@ -1085,6 +1086,14 @@ pub trait FileSystem {
         Err(io::Error::from_raw_os_error(libc::ENOSYS))
     }
 
+    /// Synchronize the filesystem containing the file referenced by `inode`.  When running with
+    /// --announce-submounts, `syncfs` is called once per submount that is to be synced.  When
+    /// running without--announce-submounts, `syncfs` is called on the root mount, but all submounts
+    /// need to be synced, too.
+    fn syncfs(&self, _ctx: Context, inode: Self::Inode) -> io::Result<()> {
+        Err(io::Error::from_raw_os_error(libc::ENOSYS))
+    }
+
     /// TODO: support this
     fn getlk(&self) -> io::Result<()> {
         Err(io::Error::from_raw_os_error(libc::ENOSYS))
