@@ -6,11 +6,13 @@ use futures::executor::{ThreadPool, ThreadPoolBuilder};
 use libc::EFD_NONBLOCK;
 use log::*;
 use passthrough::xattrmap::XattrMap;
-use std::convert::{self, TryFrom};
+use std::collections::HashSet;
+use std::convert::{self, TryFrom, TryInto};
 use std::ffi::CString;
 use std::os::unix::io::{FromRawFd, RawFd};
+use std::path::Path;
 use std::str::FromStr;
-use std::sync::{Arc, Mutex, MutexGuard, RwLock};
+use std::sync::{Arc, RwLock};
 use std::{env, error, fmt, io, process};
 
 use structopt::StructOpt;
@@ -19,7 +21,7 @@ use vhost::vhost_user::message::*;
 use vhost::vhost_user::Error::PartialMessage;
 use vhost::vhost_user::{Listener, SlaveFsCacheReq};
 use vhost_user_backend::Error::HandleRequest;
-use vhost_user_backend::{VhostUserBackendMut, VhostUserDaemon, VringMutex, VringState, VringT};
+use vhost_user_backend::{VhostUserBackend, VhostUserDaemon, VringMutex, VringState, VringT};
 use virtio_bindings::bindings::virtio_net::*;
 use virtio_bindings::bindings::virtio_ring::{
     VIRTIO_RING_F_EVENT_IDX, VIRTIO_RING_F_INDIRECT_DESC,
@@ -31,7 +33,8 @@ use virtiofsd::passthrough::{self, CachePolicy, InodeFileHandlesMode, Passthroug
 use virtiofsd::sandbox::{Sandbox, SandboxMode};
 use virtiofsd::seccomp::{enable_seccomp, SeccompAction};
 use virtiofsd::server::Server;
-use virtiofsd::Error as VhostUserFsError;
+use virtiofsd::util::write_pid_file;
+use virtiofsd::{limits, Error as VhostUserFsError};
 use vm_memory::{GuestAddressSpace, GuestMemoryAtomic, GuestMemoryLoadGuard, GuestMemoryMmap};
 use vmm_sys_util::epoll::EventSet;
 use vmm_sys_util::eventfd::EventFd;
@@ -57,12 +60,6 @@ enum Error {
     HandleEventNotEpollIn,
     /// Failed to handle unknown event.
     HandleEventUnknownEvent,
-    /// Invalid compat argument found on the command line.
-    InvalidCompatArgument,
-    /// Invalid compat value found on the command line.
-    InvalidCompatValue,
-    /// Invalid xattr map compat argument found on the command line.
-    InvalidXattrMap,
     /// Iterating through the queue failed.
     IterateQueue,
     /// No memory configured.
@@ -73,11 +70,25 @@ enum Error {
     QueueReader(VufDescriptorError),
     /// Creating a queue writer failed.
     QueueWriter(VufDescriptorError),
+    /// The unshare(CLONE_FS) call failed.
+    UnshareCloneFs(io::Error),
 }
 
 impl fmt::Display for Error {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        write!(f, "virtiofsd_error: {:?}", self)
+        use self::Error::UnshareCloneFs;
+        match self {
+            UnshareCloneFs(error) => {
+                write!(
+                    f,
+                    "The unshare(CLONE_FS) syscall failed with '{}'. \
+                    If running in a container please check that the container \
+                    runtime seccomp policy allows unshare.",
+                    error
+                )
+            }
+            _ => write!(f, "{:?}", self),
+        }
     }
 }
 
@@ -89,35 +100,6 @@ impl convert::From<Error> for io::Error {
     }
 }
 
-#[derive(Clone, Debug)]
-enum LogLevel {
-    Error,
-    Warn,
-    Info,
-    Debug,
-    Trace,
-}
-
-impl fmt::Display for LogLevel {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        write!(f, "{:?}", self)
-    }
-}
-
-impl FromStr for LogLevel {
-    type Err = &'static str;
-    fn from_str(s: &str) -> std::result::Result<Self, Self::Err> {
-        match s.to_lowercase().as_str() {
-            "error" => Ok(LogLevel::Error),
-            "warn" => Ok(LogLevel::Warn),
-            "info" => Ok(LogLevel::Info),
-            "debug" => Ok(LogLevel::Debug),
-            "trace" => Ok(LogLevel::Trace),
-            _ => Err("Unknown log level"),
-        }
-    }
-}
-
 struct VhostUserFsThread<F: FileSystem + Send + Sync + 'static> {
     mem: Option<GuestMemoryAtomic<GuestMemoryMmap>>,
     kill_evt: EventFd,
@@ -143,9 +125,25 @@ impl<F: FileSystem + Send + Sync + 'static> Clone for VhostUserFsThread<F> {
 
 impl<F: FileSystem + Send + Sync + 'static> VhostUserFsThread<F> {
     fn new(fs: F, thread_pool_size: usize) -> Result<Self> {
+        // Test that unshare(CLONE_FS) works, it will be called for each thread.
+        // It's an unprivileged system call but some Docker/Moby versions are
+        // known to reject it via seccomp when CAP_SYS_ADMIN is not given.
+        //
+        // Note that the program is single-threaded here so this syscall has no
+        // visible effect and is safe to make.
+        let ret = unsafe { libc::unshare(libc::CLONE_FS) };
+        if ret == -1 {
+            return Err(Error::UnshareCloneFs(std::io::Error::last_os_error()));
+        }
+
         let pool = if thread_pool_size > 0 {
             Some(
                 ThreadPoolBuilder::new()
+                    .after_start(|_| {
+                        // unshare FS for xattr operation
+                        let ret = unsafe { libc::unshare(libc::CLONE_FS) };
+                        assert_eq!(ret, 0); // Should not fail
+                    })
                     .pool_size(thread_pool_size)
                     .create()
                     .map_err(Error::CreateThreadPool)?,
@@ -164,7 +162,39 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserFsThread<F> {
         })
     }
 
-    fn process_queue_pool(&mut self, vring: VringMutex) -> Result<bool> {
+    fn return_descriptor(
+        vring_state: &mut VringState,
+        head_index: u16,
+        event_idx: bool,
+        len: usize,
+    ) {
+        let used_len: u32 = match len.try_into() {
+            Ok(l) => l,
+            Err(_) => panic!("Invalid used length, can't return used descritors to the ring"),
+        };
+
+        if vring_state.add_used(head_index, used_len).is_err() {
+            warn!("Couldn't return used descriptors to the ring");
+        }
+
+        if event_idx {
+            match vring_state.needs_notification() {
+                Err(_) => {
+                    warn!("Couldn't check if queue needs to be notified");
+                    vring_state.signal_used_queue().unwrap();
+                }
+                Ok(needs_notification) => {
+                    if needs_notification {
+                        vring_state.signal_used_queue().unwrap();
+                    }
+                }
+            }
+        } else {
+            vring_state.signal_used_queue().unwrap();
+        }
+    }
+
+    fn process_queue_pool(&self, vring: VringMutex) -> Result<bool> {
         let mut used_any = false;
         let atomic_mem = match &self.mem {
             Some(m) => m,
@@ -199,45 +229,25 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserFsThread<F> {
                     .map_err(Error::QueueWriter)
                     .unwrap();
 
-                server
+                let len = server
                     .handle_message(reader, writer, vu_req.as_mut())
                     .map_err(Error::ProcessQueue)
                     .unwrap();
 
-                if event_idx {
-                    if worker_vring.add_used(head_index, 0).is_err() {
-                        warn!("Couldn't return used descriptors to the ring");
-                    }
-
-                    match worker_vring.needs_notification() {
-                        Err(_) => {
-                            warn!("Couldn't check if queue needs to be notified");
-                            worker_vring.signal_used_queue().unwrap();
-                        }
-                        Ok(needs_notification) => {
-                            if needs_notification {
-                                worker_vring.signal_used_queue().unwrap();
-                            }
-                        }
-                    }
-                } else {
-                    if worker_vring.add_used(head_index, 0).is_err() {
-                        warn!("Couldn't return used descriptors to the ring");
-                    }
-                    worker_vring.signal_used_queue().unwrap();
-                }
+                Self::return_descriptor(&mut worker_vring.get_mut(), head_index, event_idx, len);
             });
         }
 
         Ok(used_any)
     }
 
-    fn process_queue_serial(&mut self, vring_state: &mut MutexGuard<VringState>) -> Result<bool> {
+    fn process_queue_serial(&self, vring_state: &mut VringState) -> Result<bool> {
         let mut used_any = false;
         let mem = match &self.mem {
             Some(m) => m.memory(),
             None => return Err(Error::NoMemoryConfigured),
         };
+        let mut vu_req = self.vu_req.clone();
 
         let avail_chains: Vec<DescriptorChain<GuestMemoryLoadGuard<GuestMemoryMmap>>> = vring_state
             .get_queue_mut()
@@ -257,40 +267,20 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserFsThread<F> {
                 .map_err(Error::QueueWriter)
                 .unwrap();
 
-            self.server
-                .handle_message(reader, writer, self.vu_req.as_mut())
+            let len = self
+                .server
+                .handle_message(reader, writer, vu_req.as_mut())
                 .map_err(Error::ProcessQueue)
                 .unwrap();
 
-            if self.event_idx {
-                if vring_state.add_used(head_index, 0).is_err() {
-                    warn!("Couldn't return used descriptors to the ring");
-                }
-
-                match vring_state.needs_notification() {
-                    Err(_) => {
-                        warn!("Couldn't check if queue needs to be notified");
-                        vring_state.signal_used_queue().unwrap();
-                    }
-                    Ok(needs_notification) => {
-                        if needs_notification {
-                            vring_state.signal_used_queue().unwrap();
-                        }
-                    }
-                }
-            } else {
-                if vring_state.add_used(head_index, 0).is_err() {
-                    warn!("Couldn't return used descriptors to the ring");
-                }
-                vring_state.signal_used_queue().unwrap();
-            }
+            Self::return_descriptor(vring_state, head_index, self.event_idx, len);
         }
 
         Ok(used_any)
     }
 
     fn handle_event_pool(
-        &mut self,
+        &self,
         device_event: u16,
         vrings: &[VringMutex],
     ) -> VhostUserBackendResult<bool> {
@@ -327,7 +317,7 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserFsThread<F> {
     }
 
     fn handle_event_serial(
-        &mut self,
+        &self,
         device_event: u16,
         vrings: &[VringMutex],
     ) -> VhostUserBackendResult<bool> {
@@ -365,19 +355,17 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserFsThread<F> {
 }
 
 struct VhostUserFsBackend<F: FileSystem + Send + Sync + 'static> {
-    thread: Mutex<VhostUserFsThread<F>>,
+    thread: RwLock<VhostUserFsThread<F>>,
 }
 
 impl<F: FileSystem + Send + Sync + 'static> VhostUserFsBackend<F> {
     fn new(fs: F, thread_pool_size: usize) -> Result<Self> {
-        let thread = Mutex::new(VhostUserFsThread::new(fs, thread_pool_size)?);
+        let thread = RwLock::new(VhostUserFsThread::new(fs, thread_pool_size)?);
         Ok(VhostUserFsBackend { thread })
     }
 }
 
-impl<F: FileSystem + Send + Sync + 'static> VhostUserBackendMut<VringMutex>
-    for VhostUserFsBackend<F>
-{
+impl<F: FileSystem + Send + Sync + 'static> VhostUserBackend<VringMutex> for VhostUserFsBackend<F> {
     fn num_queues(&self) -> usize {
         NUM_QUEUES
     }
@@ -401,20 +389,17 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserBackendMut<VringMutex>
             | VhostUserProtocolFeatures::CONFIGURE_MEM_SLOTS
     }
 
-    fn set_event_idx(&mut self, enabled: bool) {
-        self.thread.lock().unwrap().event_idx = enabled;
+    fn set_event_idx(&self, enabled: bool) {
+        self.thread.write().unwrap().event_idx = enabled;
     }
 
-    fn update_memory(
-        &mut self,
-        mem: GuestMemoryAtomic<GuestMemoryMmap>,
-    ) -> VhostUserBackendResult<()> {
-        self.thread.lock().unwrap().mem = Some(mem);
+    fn update_memory(&self, mem: GuestMemoryAtomic<GuestMemoryMmap>) -> VhostUserBackendResult<()> {
+        self.thread.write().unwrap().mem = Some(mem);
         Ok(())
     }
 
     fn handle_event(
-        &mut self,
+        &self,
         device_event: u16,
         evset: EventSet,
         vrings: &[VringMutex],
@@ -424,7 +409,7 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserBackendMut<VringMutex>
             return Err(Error::HandleEventNotEpollIn.into());
         }
 
-        let mut thread = self.thread.lock().unwrap();
+        let thread = self.thread.read().unwrap();
 
         if thread.pool.is_some() {
             thread.handle_event_pool(device_event, vrings)
@@ -434,11 +419,11 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserBackendMut<VringMutex>
     }
 
     fn exit_event(&self, _thread_index: usize) -> Option<EventFd> {
-        Some(self.thread.lock().unwrap().kill_evt.try_clone().unwrap())
+        Some(self.thread.read().unwrap().kill_evt.try_clone().unwrap())
     }
 
-    fn set_slave_req_fd(&mut self, vu_req: SlaveFsCacheReq) {
-        self.thread.lock().unwrap().vu_req = Some(vu_req);
+    fn set_slave_req_fd(&self, vu_req: SlaveFsCacheReq) {
+        self.thread.write().unwrap().vu_req = Some(vu_req);
     }
 }
 
@@ -446,11 +431,53 @@ fn parse_seccomp(src: &str) -> std::result::Result<SeccompAction, &'static str>
     Ok(match src {
         "none" => SeccompAction::Allow, // i.e. no seccomp
         "kill" => SeccompAction::Kill,
+        "log" => SeccompAction::Log,
         "trap" => SeccompAction::Trap,
         _ => return Err("Matching variant not found"),
     })
 }
 
+/// On the command line, we want to allow aliases for `InodeFileHandlesMode` values.  This enum has
+/// all values allowed on the command line, and with `From`/`Into`, it can be translated into the
+/// internally used `InodeFileHandlesMode` enum.
+#[derive(Debug, Copy, Clone, PartialEq)]
+enum InodeFileHandlesCommandLineMode {
+    /// `InodeFileHandlesMode::Never`
+    Never,
+    /// Alias for `InodeFileHandlesMode::Prefer`
+    Fallback,
+    /// `InodeFileHandlesMode::Prefer`
+    Prefer,
+    /// `InodeFileHandlesMode::Mandatory`
+    Mandatory,
+}
+
+impl From<InodeFileHandlesCommandLineMode> for InodeFileHandlesMode {
+    fn from(clm: InodeFileHandlesCommandLineMode) -> Self {
+        match clm {
+            InodeFileHandlesCommandLineMode::Never => InodeFileHandlesMode::Never,
+            InodeFileHandlesCommandLineMode::Fallback => InodeFileHandlesMode::Prefer,
+            InodeFileHandlesCommandLineMode::Prefer => InodeFileHandlesMode::Prefer,
+            InodeFileHandlesCommandLineMode::Mandatory => InodeFileHandlesMode::Mandatory,
+        }
+    }
+}
+
+impl FromStr for InodeFileHandlesCommandLineMode {
+    type Err = &'static str;
+
+    fn from_str(s: &str) -> std::result::Result<Self, Self::Err> {
+        match s {
+            "never" => Ok(InodeFileHandlesCommandLineMode::Never),
+            "fallback" => Ok(InodeFileHandlesCommandLineMode::Fallback),
+            "prefer" => Ok(InodeFileHandlesCommandLineMode::Prefer),
+            "mandatory" => Ok(InodeFileHandlesCommandLineMode::Mandatory),
+
+            _ => Err("invalid inode file handles mode"),
+        }
+    }
+}
+
 #[derive(Clone, Debug, StructOpt)]
 #[structopt(name = "virtiofsd backend", about = "Launch a virtiofsd backend.")]
 struct Opt {
@@ -458,7 +485,7 @@ struct Opt {
     #[structopt(long)]
     shared_dir: Option<String>,
 
-    /// vhost-user socket path (deprecated)
+    /// vhost-user socket path [deprecated]
     #[structopt(long, required_unless_one = &["fd", "socket-path", "print-capabilities"])]
     socket: Option<String>,
 
@@ -482,6 +509,10 @@ struct Opt {
     #[structopt(long)]
     xattr: bool,
 
+    /// Enable support for posix ACLs (implies --xattr)
+    #[structopt(long)]
+    posix_acl: bool,
+
     /// Add custom rules for translating extended attributes between host and guest
     /// (e.g. :map::user.virtiofs.:)
     #[structopt(long, parse(try_from_str = <XattrMap as TryFrom<&str>>::try_from))]
@@ -491,7 +522,7 @@ struct Opt {
     #[structopt(long, default_value = "namespace")]
     sandbox: SandboxMode,
 
-    /// Action to take when seccomp finds a not allowed syscall (allow, kill, log, trap)
+    /// Action to take when seccomp finds a not allowed syscall (none, kill, log, trap)
     #[structopt(long, parse(try_from_str = parse_seccomp), default_value = "kill")]
     seccomp: SeccompAction,
 
@@ -500,19 +531,23 @@ struct Opt {
     announce_submounts: bool,
 
     /// When to use file handles to reference inodes instead of O_PATH file descriptors (never,
-    /// fallback)
+    /// prefer, mandatory)
     ///
     /// - never: Never use file handles, always use O_PATH file descriptors.
     ///
-    /// - fallback: Attempt to generate file handles, but fall back to O_PATH file descriptors
-    /// where the underlying filesystem does not support file handles.
+    /// - prefer: Attempt to generate file handles, but fall back to O_PATH file descriptors where
+    /// the underlying filesystem does not support file handles.  Useful when there are various
+    /// different filesystems under the shared directory and some of them do not support file
+    /// handles.  ("fallback" is a deprecated alias for "prefer".)
+    ///
+    /// - mandatory: Always use file handles, never fall back to O_PATH file descriptors.
     ///
     /// Using file handles reduces the number of file descriptors virtiofsd keeps open, which is
     /// not only helpful with resources, but may also be important in cases where virtiofsd should
     /// only have file descriptors open for files that are open in the guest, e.g. to get around
     /// bad interactions with NFS's silly renaming.
     #[structopt(long, require_equals = true, default_value = "never")]
-    inode_file_handles: InodeFileHandlesMode,
+    inode_file_handles: InodeFileHandlesCommandLineMode,
 
     /// The caching policy the file system should use (auto, always, never)
     #[structopt(long, default_value = "auto")]
@@ -534,54 +569,89 @@ struct Opt {
     #[structopt(long = "print-capabilities")]
     print_capabilities: bool,
 
-    /// Log level (error, warn, info, debug, trace)
-    #[structopt(long = "log-level", default_value = "info")]
-    log_level: LogLevel,
+    /// Modify the list of capabilities, e.g., --modcaps=+sys_admin:-chown
+    #[structopt(long)]
+    modcaps: Option<String>,
 
-    /// Set maximum number of file descriptors (0 leaves rlimit unchanged) [default: the value read from `/proc/sys/fs/nr_open`]
+    /// Log level (error, warn, info, debug, trace, off)
+    #[structopt(long = "log-level", default_value = "info")]
+    log_level: LevelFilter,
+
+    /// Log to syslog [default: stderr]
+    #[structopt(long)]
+    syslog: bool,
+
+    /// Set maximum number of file descriptors (0 leaves rlimit unchanged)
+    /// [default: min(1000000, '/proc/sys/fs/nr_open')]
     #[structopt(long = "rlimit-nofile")]
     rlimit_nofile: Option<u64>,
 
-    /// Options in a format compatible with the legacy implementation
+    /// Options in a format compatible with the legacy implementation [deprecated]
     #[structopt(short = "o")]
     compat_options: Option<Vec<String>>,
 
-    /// Set log level to "debug" in a compatible way with the legacy implementation
+    /// Set log level to "debug" [deprecated]
     #[structopt(short = "d")]
     compat_debug: bool,
+
+    /// Disable KILLPRIV V2 support
+    #[structopt(long)]
+    no_killpriv_v2: bool,
+
+    /// Compatibility option that has no effect [deprecated]
+    #[structopt(short = "f")]
+    compat_foreground: bool,
 }
 
-fn parse_compat(opt: Opt) -> Result<Opt> {
-    fn parse_tuple(opt: &mut Opt, tuple: &str) -> Result<()> {
+fn parse_compat(opt: Opt) -> Opt {
+    use structopt::clap::{Error, ErrorKind};
+    fn value_error(arg: &str, value: &str) -> ! {
+        Error::with_description(
+            format!("Invalid compat value '{}' for '-o {}'", value, arg).as_str(),
+            ErrorKind::InvalidValue,
+        )
+        .exit()
+    }
+    fn argument_error(arg: &str) -> ! {
+        Error::with_description(
+            format!("Invalid compat argument '-o {}'", arg).as_str(),
+            ErrorKind::UnknownArgument,
+        )
+        .exit()
+    }
+
+    fn parse_tuple(opt: &mut Opt, tuple: &str) {
         match tuple.split('=').collect::<Vec<&str>>()[..] {
             ["xattrmap", value] => {
-                opt.xattrmap = Some(XattrMap::try_from(value).map_err(|_| Error::InvalidXattrMap)?)
+                opt.xattrmap = Some(
+                    XattrMap::try_from(value).unwrap_or_else(|_| value_error("xattrmap", value)),
+                )
             }
             ["cache", value] => match value {
                 "auto" => opt.cache = CachePolicy::Auto,
                 "always" => opt.cache = CachePolicy::Always,
                 "none" => opt.cache = CachePolicy::Never,
-                _ => return Err(Error::InvalidCompatValue),
+                _ => value_error("cache", value),
             },
             ["loglevel", value] => match value {
-                "debug" => opt.log_level = LogLevel::Debug,
-                "info" => opt.log_level = LogLevel::Info,
-                "warn" => opt.log_level = LogLevel::Warn,
-                "err" => opt.log_level = LogLevel::Error,
-                _ => return Err(Error::InvalidCompatValue),
+                "debug" => opt.log_level = LevelFilter::Debug,
+                "info" => opt.log_level = LevelFilter::Info,
+                "warn" => opt.log_level = LevelFilter::Warn,
+                "err" => opt.log_level = LevelFilter::Error,
+                _ => value_error("loglevel", value),
             },
             ["sandbox", value] => match value {
                 "namespace" => opt.sandbox = SandboxMode::Namespace,
                 "chroot" => opt.sandbox = SandboxMode::Chroot,
-                _ => return Err(Error::InvalidCompatValue),
+                _ => value_error("sandbox", value),
             },
             ["source", value] => opt.shared_dir = Some(value.to_string()),
-            _ => return Err(Error::InvalidCompatArgument),
+            ["modcaps", value] => opt.modcaps = Some(value.to_string()),
+            _ => argument_error(tuple),
         }
-        Ok(())
     }
 
-    fn parse_single(opt: &mut Opt, option: &str) -> Result<()> {
+    fn parse_single(opt: &mut Opt, option: &str) {
         match option {
             "xattr" => opt.xattr = true,
             "no_xattr" => opt.xattr = false,
@@ -592,9 +662,13 @@ fn parse_compat(opt: Opt) -> Result<Opt> {
             "allow_direct_io" => opt.allow_direct_io = true,
             "no_allow_direct_io" => opt.allow_direct_io = false,
             "announce_submounts" => opt.announce_submounts = true,
-            _ => return Err(Error::InvalidCompatArgument),
+            "killpriv_v2" => opt.no_killpriv_v2 = false,
+            "no_killpriv_v2" => opt.no_killpriv_v2 = true,
+            "posix_acl" => opt.posix_acl = true,
+            "no_posix_acl" => opt.posix_acl = false,
+            "no_posix_lock" | "no_flock" => (),
+            _ => argument_error(option),
         }
-        Ok(())
     }
 
     let mut clean_opt = opt.clone();
@@ -603,15 +677,15 @@ fn parse_compat(opt: Opt) -> Result<Opt> {
         for line in compat_options {
             for option in line.to_string().split(',') {
                 if option.contains('=') {
-                    parse_tuple(&mut clean_opt, option)?;
+                    parse_tuple(&mut clean_opt, option);
                 } else {
-                    parse_single(&mut clean_opt, option)?;
+                    parse_single(&mut clean_opt, option);
                 }
             }
         }
     }
 
-    Ok(clean_opt)
+    clean_opt
 }
 
 fn print_capabilities() {
@@ -620,14 +694,88 @@ fn print_capabilities() {
     println!("}}");
 }
 
-fn initialize_logging(log_level: &LogLevel) {
-    match env::var("RUST_LOG") {
-        Ok(_) => {}
-        Err(_) => env::set_var("RUST_LOG", log_level.to_string()),
+fn set_default_logger(log_level: LevelFilter) {
+    if env::var("RUST_LOG").is_err() {
+        env::set_var("RUST_LOG", log_level.to_string());
     }
     env_logger::init();
 }
 
+fn initialize_logging(opt: &Opt) {
+    let log_level = if opt.compat_debug {
+        LevelFilter::Debug
+    } else {
+        opt.log_level
+    };
+
+    if opt.syslog {
+        if let Err(e) = syslog::init(syslog::Facility::LOG_USER, log_level, None) {
+            set_default_logger(log_level);
+            warn!("can't enable syslog: {}", e);
+        }
+    } else {
+        set_default_logger(log_level);
+    }
+}
+
+fn set_signal_handlers() {
+    use vmm_sys_util::signal;
+
+    extern "C" fn handle_signal(_: libc::c_int, _: *mut libc::siginfo_t, _: *mut libc::c_void) {
+        unsafe { libc::_exit(1) };
+    }
+    let signals = vec![libc::SIGHUP, libc::SIGTERM];
+    for s in signals {
+        if let Err(e) = signal::register_signal_handler(s, handle_signal) {
+            error!("Setting signal handlers: {}", e);
+            process::exit(1);
+        }
+    }
+}
+
+fn parse_modcaps(
+    default_caps: Vec<&str>,
+    modcaps: Option<String>,
+) -> (HashSet<String>, HashSet<String>) {
+    let mut required_caps: HashSet<String> = default_caps.iter().map(|&s| s.into()).collect();
+    let mut disabled_caps = HashSet::new();
+
+    if let Some(modcaps) = modcaps {
+        for modcap in modcaps.split(':').map(str::to_string) {
+            if modcap.is_empty() {
+                error!("empty modcap found: expected (+|-)capability:...");
+                process::exit(1);
+            }
+            let (action, cap_name) = modcap.split_at(1);
+            let cap_name = cap_name.to_uppercase();
+            if !matches!(action, "+" | "-") {
+                error!(
+                    "invalid modcap action: expecting '+'|'-' but found '{}'",
+                    action
+                );
+                process::exit(1);
+            }
+            if let Err(error) = capng::name_to_capability(&cap_name) {
+                error!("invalid capability '{}': {}", &cap_name, error);
+                process::exit(1);
+            }
+
+            match action {
+                "+" => {
+                    disabled_caps.remove(&cap_name);
+                    required_caps.insert(cap_name);
+                }
+                "-" => {
+                    required_caps.remove(&cap_name);
+                    disabled_caps.insert(cap_name);
+                }
+                _ => unreachable!(),
+            }
+        }
+    }
+    (required_caps, disabled_caps)
+}
+
 fn drop_parent_capabilities() {
     // The parent doesn't require any capabilities, as it'd be just waiting for
     // the child to exit.
@@ -638,8 +786,8 @@ fn drop_parent_capabilities() {
     }
 }
 
-fn drop_child_capabilities(inode_file_handles: InodeFileHandlesMode) {
-    let mut required_caps = vec![
+fn drop_child_capabilities(inode_file_handles: InodeFileHandlesMode, modcaps: Option<String>) {
+    let default_caps = vec![
         "CHOWN",
         "DAC_OVERRIDE",
         "FOWNER",
@@ -649,9 +797,18 @@ fn drop_child_capabilities(inode_file_handles: InodeFileHandlesMode) {
         "MKNOD",
         "SETFCAP",
     ];
+    let (mut required_caps, disabled_caps) = parse_modcaps(default_caps, modcaps);
 
     if inode_file_handles != InodeFileHandlesMode::Never {
-        required_caps.push("DAC_READ_SEARCH");
+        let required_cap = "DAC_READ_SEARCH".to_owned();
+        if disabled_caps.contains(&required_cap) {
+            error!(
+                "can't disable {} when using --inode-file-handles={:?}",
+                &required_cap, inode_file_handles
+            );
+            process::exit(1);
+        }
+        required_caps.insert(required_cap);
     }
 
     capng::clear(capng::Set::BOTH);
@@ -660,7 +817,7 @@ fn drop_child_capabilities(inode_file_handles: InodeFileHandlesMode) {
     if let Err(e) = capng::updatev(
         capng::Action::ADD,
         capng::Type::PERMITTED | capng::Type::EFFECTIVE,
-        required_caps,
+        required_caps.iter().map(String::as_str).collect(),
     ) {
         error!("can't set up the child capabilities: {}", e);
         process::exit(1);
@@ -672,18 +829,17 @@ fn drop_child_capabilities(inode_file_handles: InodeFileHandlesMode) {
 }
 
 fn main() {
-    let opt = parse_compat(Opt::from_args()).expect("invalid compat argument");
+    let opt = parse_compat(Opt::from_args());
+
+    let killpriv_v2 = !opt.no_killpriv_v2;
 
     if opt.print_capabilities {
         print_capabilities();
         return;
     }
 
-    if opt.compat_debug {
-        initialize_logging(&LogLevel::Debug)
-    } else {
-        initialize_logging(&opt.log_level)
-    }
+    initialize_logging(&opt);
+    set_signal_handlers();
 
     let shared_dir = match opt.shared_dir.as_ref() {
         Some(s) => s,
@@ -692,9 +848,22 @@ fn main() {
             process::exit(1);
         }
     };
+    if opt.compat_foreground {
+        warn!("Use of deprecated flag '-f': This flag has no effect, please remove it");
+    }
+    if opt.compat_debug {
+        warn!("Use of deprecated flag '-d': Please use the '--log-level debug' option instead");
+    }
+    if opt.compat_options.is_some() {
+        warn!("Use of deprecated option format '-o': Please specify options without it (e.g., '--cache auto' instead of '-o cache=auto')");
+    }
+    if opt.inode_file_handles == InodeFileHandlesCommandLineMode::Fallback {
+        warn!("Use of deprecated value 'fallback' for '--inode-file-handles': Please use 'prefer' instead");
+    }
+
     let sandbox_mode = opt.sandbox.clone();
     let xattrmap = opt.xattrmap.clone();
-    let xattr = if xattrmap.is_some() { true } else { opt.xattr };
+    let xattr = xattrmap.is_some() || opt.posix_acl || opt.xattr;
     let thread_pool_size = opt.thread_pool_size;
     let readdirplus = match opt.cache {
         CachePolicy::Never => false,
@@ -712,16 +881,26 @@ fn main() {
             | libc::S_IXOTH
     };
 
-    let (listener, socket_path) = match opt.fd.as_ref() {
-        Some(fd) => unsafe { (Listener::from_raw_fd(*fd), None) },
+    // We need to keep _pid_file around because it maintains a lock on the pid file
+    // that prevents another daemon from using the same pid file.
+    let (listener, socket_path, _pid_file) = match opt.fd.as_ref() {
+        Some(fd) => unsafe { (Listener::from_raw_fd(*fd), None, None) },
         None => {
             // Set umask to ensure the socket is created with the right permissions
             let old_umask = unsafe { libc::umask(umask) };
 
             let socket = opt.socket_path.as_ref().unwrap_or_else(|| {
-                println!("warning: use of deprecated parameter '--socket': Please use the '--socket-path' option instead.");
+                warn!("use of deprecated parameter '--socket': Please use the '--socket-path' option instead");
                 opt.socket.as_ref().unwrap() // safe to unwrap because clap ensures either --socket or --sock are passed
             });
+
+            let pid_file_name = socket.to_owned() + ".pid";
+            let pid_file_path = Path::new(pid_file_name.as_str());
+            let pid_file = write_pid_file(pid_file_path).unwrap_or_else(|error| {
+                error!("Error creating pid file '{}': {}", pid_file_name, error);
+                process::exit(1);
+            });
+
             let listener = Listener::new(socket, true).unwrap_or_else(|error| {
                 error!("Error creating listener: {}", error);
                 process::exit(1);
@@ -730,7 +909,7 @@ fn main() {
             // Restore umask
             unsafe { libc::umask(old_umask) };
 
-            (listener, Some(socket.clone()))
+            (listener, Some(socket.clone()), Some(pid_file))
         }
     };
 
@@ -754,30 +933,44 @@ fn main() {
         }
     }
 
-    let rlimit_nofile = if let Some(rlimit_nofile) = opt.rlimit_nofile {
-        if rlimit_nofile != 0 {
-            Some(rlimit_nofile)
-        } else {
-            None
-        }
-    } else {
-        None
-    };
+    limits::setup_rlimit_nofile(opt.rlimit_nofile).unwrap_or_else(|error| {
+        error!("Error increasing number of open files: {}", error);
+        process::exit(1)
+    });
 
-    let mut sandbox = Sandbox::new(shared_dir.to_string(), sandbox_mode, rlimit_nofile)
-        .unwrap_or_else(|error| {
-            error!("Error creating sandbox: {}", error);
-            process::exit(1)
-        });
+    let mut sandbox = Sandbox::new(shared_dir.to_string(), sandbox_mode).unwrap_or_else(|error| {
+        error!("Error creating sandbox: {}", error);
+        process::exit(1)
+    });
     let fs_cfg = match sandbox.enter().unwrap_or_else(|error| {
         error!("Error entering sandbox: {}", error);
         process::exit(1)
     }) {
+        // `enter()` returns the PID of the child to the parent, if it forked.
         Some(child_pid) => {
             drop_parent_capabilities();
-            unsafe { libc::waitpid(child_pid, std::ptr::null_mut(), 0) };
-            return;
+            let mut status = 0;
+            // On success, `libc::waitpid()` returns the PID of the child.
+            if unsafe { libc::waitpid(child_pid, &mut status, 0) } != child_pid {
+                error!("Error during waitpid()");
+                process::exit(1);
+            }
+
+            let exit_code = if libc::WIFEXITED(status) {
+                libc::WEXITSTATUS(status)
+            } else if libc::WIFSIGNALED(status) {
+                let signal = libc::WTERMSIG(status);
+                error!("Child process terminated by signal {}", signal);
+                -signal
+            } else {
+                error!("Unexpected waitpid status: {:#X}", status);
+                libc::EXIT_FAILURE
+            };
+
+            process::exit(exit_code);
         }
+        // `enter()` returns `None` to the process that should proceed (i.e. to the child, if it
+        // forked).
         None => passthrough::Config {
             cache_policy: opt.cache,
             root_dir: sandbox.get_root_dir(),
@@ -787,10 +980,12 @@ fn main() {
             proc_sfd_rawfd: sandbox.get_proc_self_fd(),
             proc_mountinfo_rawfd: sandbox.get_mountinfo_fd(),
             announce_submounts: opt.announce_submounts,
-            inode_file_handles: opt.inode_file_handles,
+            inode_file_handles: opt.inode_file_handles.into(),
             readdirplus,
             writeback: opt.writeback,
             allow_direct_io: opt.allow_direct_io,
+            killpriv_v2,
+            posix_acl: opt.posix_acl,
             ..Default::default()
         },
     };
@@ -798,15 +993,28 @@ fn main() {
     // Must happen before we start the thread pool
     match opt.seccomp {
         SeccompAction::Allow => {}
-        _ => enable_seccomp(opt.seccomp).unwrap(),
+        _ => enable_seccomp(opt.seccomp, opt.syslog).unwrap(),
     }
 
-    drop_child_capabilities(opt.inode_file_handles);
+    drop_child_capabilities(fs_cfg.inode_file_handles, opt.modcaps);
 
-    let fs = PassthroughFs::new(fs_cfg).unwrap();
-    let fs_backend = Arc::new(RwLock::new(
-        VhostUserFsBackend::new(fs, thread_pool_size).unwrap(),
-    ));
+    let fs = match PassthroughFs::new(fs_cfg) {
+        Ok(fs) => fs,
+        Err(e) => {
+            error!(
+                "Failed to create internal filesystem representation: {:?}",
+                e
+            );
+            process::exit(1);
+        }
+    };
+
+    let fs_backend = Arc::new(
+        VhostUserFsBackend::new(fs, thread_pool_size).unwrap_or_else(|error| {
+            error!("Error creating vhost-user backend: {}", error);
+            process::exit(1)
+        }),
+    );
 
     let mut daemon = VhostUserDaemon::new(
         String::from("virtiofsd-backend"),
@@ -832,10 +1040,8 @@ fn main() {
     }
 
     let kill_evt = fs_backend
-        .read()
-        .unwrap()
         .thread
-        .lock()
+        .read()
         .unwrap()
         .kill_evt
         .try_clone()
