@@ -28,8 +28,12 @@ pub enum Error {
     CleanMount(io::Error),
     /// Failed to create a temporary directory.
     CreateTempDir(io::Error),
+    /// Failed to drop supplemental groups.
+    DropSupplementalGroups(io::Error),
     /// Call to libc::fork returned an error.
     Fork(io::Error),
+    /// Failed to get the number of supplemental groups.
+    GetSupplementalGroups(io::Error),
     /// Error bind-mounting a directory.
     MountBind(io::Error),
     /// Failed to mount old root.
@@ -60,12 +64,12 @@ pub enum Error {
     UmountTempDir(io::Error),
     /// Call to libc::unshare returned an error.
     Unshare(io::Error),
-    /// Failed to read from procfs.
-    ReadProc(io::Error),
-    /// Failed to parse `/proc/sys/fs/nr_open`.
-    InvalidNrOpen(std::num::ParseIntError),
-    /// Failed to set rlimit.
-    SetRlimit(io::Error),
+    /// Failed to write to `/proc/self/gid_map`.
+    WriteGidMap(io::Error),
+    /// Failed to write to `/proc/self/setgroups`.
+    WriteSetGroups(io::Error),
+    /// Failed to write to `/proc/self/uid_map`.
+    WriteUidMap(io::Error),
     /// Sandbox mode unavailable for non-privileged users
     SandboxModeInvalidUID,
 }
@@ -120,16 +124,10 @@ pub struct Sandbox {
     mountinfo_fd: Option<File>,
     /// Mechanism to be used for setting up the sandbox.
     sandbox_mode: SandboxMode,
-    /// Value to set as RLIMIT_NOFILE
-    rlimit_nofile: Option<u64>,
 }
 
 impl Sandbox {
-    pub fn new(
-        shared_dir: String,
-        sandbox_mode: SandboxMode,
-        rlimit_nofile: Option<u64>,
-    ) -> io::Result<Self> {
+    pub fn new(shared_dir: String, sandbox_mode: SandboxMode) -> io::Result<Self> {
         let shared_dir_rp = fs::canonicalize(shared_dir)?;
         let shared_dir_rp_str = shared_dir_rp
             .to_str()
@@ -140,7 +138,6 @@ impl Sandbox {
             proc_self_fd: None,
             mountinfo_fd: None,
             sandbox_mode,
-            rlimit_nofile,
         })
     }
 
@@ -342,32 +339,23 @@ impl Sandbox {
         Ok(())
     }
 
-    /// Sets the limit of open files to the max possible.
-    fn setup_nofile_rlimit(&self) -> Result<(), Error> {
-        let rlimit_nofile = if let Some(rlimit_nofile) = self.rlimit_nofile {
-            rlimit_nofile
-        } else {
-            // /proc/sys/fs/nr_open is a sysctl file that shows the maximum number
-            // of file-handles a process can allocate.
-            let path = "/proc/sys/fs/nr_open";
-            let max_str = fs::read_to_string(path).map_err(Error::ReadProc)?;
-            max_str.trim().parse().map_err(Error::InvalidNrOpen)?
-        };
+    /// Sets 1-to-1 mappings for the current uid and gid.
+    fn setup_id_mappings(&self, uid: u32, gid: u32) -> Result<(), Error> {
+        // To be able to set up the gid mapping, we're required to disable setgroups(2) first.
+        fs::write("/proc/self/setgroups", "deny\n").map_err(Error::WriteSetGroups)?;
 
-        let limit = libc::rlimit {
-            rlim_cur: rlimit_nofile,
-            rlim_max: rlimit_nofile,
-        };
-        let ret = unsafe { libc::setrlimit(libc::RLIMIT_NOFILE, &limit) };
-        if ret < 0 {
-            Err(Error::SetRlimit(std::io::Error::last_os_error()))
-        } else {
-            Ok(())
-        }
+        // Set up 1-to-1 mappings for our uid and gid.
+        let uid_mapping = format!("{} {} 1\n", uid, uid);
+        fs::write("/proc/self/uid_map", uid_mapping).map_err(Error::WriteUidMap)?;
+
+        let gid_mapping = format!("{} {} 1\n", gid, gid);
+        fs::write("/proc/self/gid_map", gid_mapping).map_err(Error::WriteGidMap)?;
+        Ok(())
     }
 
     pub fn enter_namespace(&mut self) -> Result<Option<i32>, Error> {
         let uid = unsafe { libc::geteuid() };
+        let gid = unsafe { libc::getegid() };
 
         let flags = if uid == 0 {
             libc::CLONE_NEWPID | libc::CLONE_NEWNS | libc::CLONE_NEWNET
@@ -378,6 +366,21 @@ impl Sandbox {
             libc::CLONE_NEWPID | libc::CLONE_NEWNS | libc::CLONE_NEWNET | libc::CLONE_NEWUSER
         };
 
+        // Drop supplemental groups. This is running as root and will
+        // support arbitrary uid/gid switching and we don't want to
+        // retain membership of any supplementary groups.
+        //
+        // This is not necessarily required for non-root case, where
+        // unprivileged user has started us, we will setup one user
+        // namespace with 1:1 mapping and there is no arbitrary uid/gid
+        // switching at all. In this mode setgroups() is not allowed, so
+        // we can't drop supplementary groups even if wanted to. Only
+        // way to do this will be to use newuidmap/newgidmap to setup
+        // user namespace which will allow setgroups().
+        if uid == 0 {
+            self.drop_supplemental_groups()?;
+        }
+
         let ret = unsafe { libc::unshare(flags) };
         if ret != 0 {
             return Err(Error::Unshare(std::io::Error::last_os_error()));
@@ -387,9 +390,10 @@ impl Sandbox {
         match child {
             0 => {
                 // This is the child. Request to receive SIGTERM on parent's death.
+                // FIXME: Race condition: the parent process might have died already.
                 unsafe { libc::prctl(libc::PR_SET_PDEATHSIG, libc::SIGTERM) };
-                if uid == 0 {
-                    self.setup_nofile_rlimit()?;
+                if uid != 0 {
+                    self.setup_id_mappings(uid, gid)?;
                 }
                 self.setup_mounts()?;
                 Ok(None)
@@ -434,6 +438,22 @@ impl Sandbox {
         Ok(None)
     }
 
+    fn drop_supplemental_groups(&self) -> Result<(), Error> {
+        let ngroups = unsafe { libc::getgroups(0, std::ptr::null_mut()) };
+        if ngroups < 0 {
+            return Err(Error::GetSupplementalGroups(std::io::Error::last_os_error()));
+        } else if ngroups != 0 {
+            let ret = unsafe { libc::setgroups(0, std::ptr::null()) };
+            if ret != 0 {
+                return Err(Error::DropSupplementalGroups(
+                    std::io::Error::last_os_error(),
+                ));
+            }
+        }
+
+        Ok(())
+    }
+
     /// Set up sandbox, fork and jump into it.
     ///
     /// On success, the returned value will be the PID of the child for the parent and `None` for
@@ -444,6 +464,10 @@ impl Sandbox {
             return Err(Error::SandboxModeInvalidUID);
         }
 
+        // Unconditionally drop supplemental groups for every sandbox mode.
+        if self.sandbox_mode != SandboxMode::Namespace {
+            self.drop_supplemental_groups()?;
+        }
         match self.sandbox_mode {
             SandboxMode::Namespace => self.enter_namespace(),
             SandboxMode::Chroot => self.enter_chroot(),
