@@ -10,11 +10,12 @@ use crate::filesystem::{
 };
 use crate::fuse::*;
 use crate::{Error, Result};
-use std::convert::TryInto;
+use std::convert::{TryFrom, TryInto};
 use std::ffi::CStr;
 use std::fs::File;
 use std::io::{self, Read, Write};
 use std::mem::{size_of, MaybeUninit};
+use std::sync::atomic::{AtomicU64, Ordering};
 use std::time::Duration;
 use vm_memory::ByteValued;
 
@@ -59,11 +60,15 @@ impl<'a> io::Write for ZcWriter<'a> {
 
 pub struct Server<F: FileSystem + Sync> {
     fs: F,
+    options: AtomicU64,
 }
 
 impl<F: FileSystem + Sync> Server<F> {
     pub fn new(fs: F) -> Server<F> {
-        Server { fs }
+        Server {
+            fs,
+            options: AtomicU64::new(FsOptions::empty().bits()),
+        }
     }
 
     #[allow(clippy::cognitive_complexity)]
@@ -83,61 +88,71 @@ impl<F: FileSystem + Sync> Server<F> {
             );
         }
 
-        debug!("Received request: {:?}", in_header.opcode);
-
-        match in_header.opcode {
-            x if x == Opcode::Lookup as u32 => self.lookup(in_header, r, w),
-            x if x == Opcode::Forget as u32 => self.forget(in_header, r), // No reply.
-            x if x == Opcode::Getattr as u32 => self.getattr(in_header, r, w),
-            x if x == Opcode::Setattr as u32 => self.setattr(in_header, r, w),
-            x if x == Opcode::Readlink as u32 => self.readlink(in_header, w),
-            x if x == Opcode::Symlink as u32 => self.symlink(in_header, r, w),
-            x if x == Opcode::Mknod as u32 => self.mknod(in_header, r, w),
-            x if x == Opcode::Mkdir as u32 => self.mkdir(in_header, r, w),
-            x if x == Opcode::Unlink as u32 => self.unlink(in_header, r, w),
-            x if x == Opcode::Rmdir as u32 => self.rmdir(in_header, r, w),
-            x if x == Opcode::Rename as u32 => self.rename(in_header, r, w),
-            x if x == Opcode::Link as u32 => self.link(in_header, r, w),
-            x if x == Opcode::Open as u32 => self.open(in_header, r, w),
-            x if x == Opcode::Read as u32 => self.read(in_header, r, w),
-            x if x == Opcode::Write as u32 => self.write(in_header, r, w),
-            x if x == Opcode::Statfs as u32 => self.statfs(in_header, w),
-            x if x == Opcode::Release as u32 => self.release(in_header, r, w),
-            x if x == Opcode::Fsync as u32 => self.fsync(in_header, r, w),
-            x if x == Opcode::Setxattr as u32 => self.setxattr(in_header, r, w),
-            x if x == Opcode::Getxattr as u32 => self.getxattr(in_header, r, w),
-            x if x == Opcode::Listxattr as u32 => self.listxattr(in_header, r, w),
-            x if x == Opcode::Removexattr as u32 => self.removexattr(in_header, r, w),
-            x if x == Opcode::Flush as u32 => self.flush(in_header, r, w),
-            x if x == Opcode::Init as u32 => self.init(in_header, r, w),
-            x if x == Opcode::Opendir as u32 => self.opendir(in_header, r, w),
-            x if x == Opcode::Readdir as u32 => self.readdir(in_header, r, w),
-            x if x == Opcode::Releasedir as u32 => self.releasedir(in_header, r, w),
-            x if x == Opcode::Fsyncdir as u32 => self.fsyncdir(in_header, r, w),
-            x if x == Opcode::Getlk as u32 => self.getlk(in_header, r, w),
-            x if x == Opcode::Setlk as u32 => self.setlk(in_header, r, w),
-            x if x == Opcode::Setlkw as u32 => self.setlkw(in_header, r, w),
-            x if x == Opcode::Access as u32 => self.access(in_header, r, w),
-            x if x == Opcode::Create as u32 => self.create(in_header, r, w),
-            x if x == Opcode::Interrupt as u32 => Ok(self.interrupt(in_header)),
-            x if x == Opcode::Bmap as u32 => self.bmap(in_header, r, w),
-            x if x == Opcode::Destroy as u32 => Ok(self.destroy()),
-            x if x == Opcode::Ioctl as u32 => self.ioctl(in_header, r, w),
-            x if x == Opcode::Poll as u32 => self.poll(in_header, r, w),
-            x if x == Opcode::NotifyReply as u32 => self.notify_reply(in_header, r, w),
-            x if x == Opcode::BatchForget as u32 => self.batch_forget(in_header, r, w),
-            x if x == Opcode::Fallocate as u32 => self.fallocate(in_header, r, w),
-            x if x == Opcode::Readdirplus as u32 => self.readdirplus(in_header, r, w),
-            x if x == Opcode::Rename2 as u32 => self.rename2(in_header, r, w),
-            x if x == Opcode::Lseek as u32 => self.lseek(in_header, r, w),
-            x if x == Opcode::CopyFileRange as u32 => self.copyfilerange(in_header, r, w),
-            x if x == Opcode::SetupMapping as u32 => self.setupmapping(in_header, r, w, vu_req),
-            x if x == Opcode::RemoveMapping as u32 => self.removemapping(in_header, r, w, vu_req),
-            _ => reply_error(
+        if let Ok(opcode) = Opcode::try_from(in_header.opcode) {
+            debug!(
+                "Received request: opcode={:?} ({}), inode={}, unique={}, pid={}",
+                opcode, in_header.opcode, in_header.nodeid, in_header.unique, in_header.pid
+            );
+            match opcode {
+                Opcode::Lookup => self.lookup(in_header, r, w),
+                Opcode::Forget => self.forget(in_header, r), // No reply.
+                Opcode::Getattr => self.getattr(in_header, r, w),
+                Opcode::Setattr => self.setattr(in_header, r, w),
+                Opcode::Readlink => self.readlink(in_header, w),
+                Opcode::Symlink => self.symlink(in_header, r, w),
+                Opcode::Mknod => self.mknod(in_header, r, w),
+                Opcode::Mkdir => self.mkdir(in_header, r, w),
+                Opcode::Unlink => self.unlink(in_header, r, w),
+                Opcode::Rmdir => self.rmdir(in_header, r, w),
+                Opcode::Rename => self.rename(in_header, r, w),
+                Opcode::Link => self.link(in_header, r, w),
+                Opcode::Open => self.open(in_header, r, w),
+                Opcode::Read => self.read(in_header, r, w),
+                Opcode::Write => self.write(in_header, r, w),
+                Opcode::Statfs => self.statfs(in_header, w),
+                Opcode::Release => self.release(in_header, r, w),
+                Opcode::Fsync => self.fsync(in_header, r, w),
+                Opcode::Setxattr => self.setxattr(in_header, r, w),
+                Opcode::Getxattr => self.getxattr(in_header, r, w),
+                Opcode::Listxattr => self.listxattr(in_header, r, w),
+                Opcode::Removexattr => self.removexattr(in_header, r, w),
+                Opcode::Flush => self.flush(in_header, r, w),
+                Opcode::Init => self.init(in_header, r, w),
+                Opcode::Opendir => self.opendir(in_header, r, w),
+                Opcode::Readdir => self.readdir(in_header, r, w),
+                Opcode::Releasedir => self.releasedir(in_header, r, w),
+                Opcode::Fsyncdir => self.fsyncdir(in_header, r, w),
+                Opcode::Getlk => self.getlk(in_header, r, w),
+                Opcode::Setlk => self.setlk(in_header, r, w),
+                Opcode::Setlkw => self.setlkw(in_header, r, w),
+                Opcode::Access => self.access(in_header, r, w),
+                Opcode::Create => self.create(in_header, r, w),
+                Opcode::Interrupt => Ok(self.interrupt(in_header)),
+                Opcode::Bmap => self.bmap(in_header, r, w),
+                Opcode::Destroy => Ok(self.destroy()),
+                Opcode::Ioctl => self.ioctl(in_header, r, w),
+                Opcode::Poll => self.poll(in_header, r, w),
+                Opcode::NotifyReply => self.notify_reply(in_header, r, w),
+                Opcode::BatchForget => self.batch_forget(in_header, r, w),
+                Opcode::Fallocate => self.fallocate(in_header, r, w),
+                Opcode::Readdirplus => self.readdirplus(in_header, r, w),
+                Opcode::Rename2 => self.rename2(in_header, r, w),
+                Opcode::Lseek => self.lseek(in_header, r, w),
+                Opcode::CopyFileRange => self.copyfilerange(in_header, r, w),
+                Opcode::SetupMapping => self.setupmapping(in_header, r, w, vu_req),
+                Opcode::RemoveMapping => self.removemapping(in_header, r, w, vu_req),
+                Opcode::Syncfs => self.syncfs(in_header, w),
+            }
+        } else {
+            debug!(
+                "Received unknown request: opcode={}, inode={}",
+                in_header.opcode, in_header.nodeid
+            );
+            reply_error(
                 io::Error::from_raw_os_error(libc::ENOSYS),
                 in_header.unique,
                 w,
-            ),
+            )
         }
     }
 
@@ -734,12 +749,37 @@ impl<F: FileSystem + Sync> Server<F> {
     }
 
     fn setxattr(&self, in_header: InHeader, mut r: Reader, w: Writer) -> Result<usize> {
-        let SetxattrIn { size, flags } = r.read_obj().map_err(Error::DecodeMessage)?;
+        let options = FsOptions::from_bits_truncate(self.options.load(Ordering::Relaxed));
+        let (
+            SetxattrIn {
+                size,
+                flags,
+                setxattr_flags,
+                ..
+            },
+            setxattrin_size,
+        ) = if options.contains(FsOptions::SETXATTR_EXT) {
+            (
+                r.read_obj().map_err(Error::DecodeMessage)?,
+                size_of::<SetxattrIn>(),
+            )
+        } else {
+            let SetxattrInCompat { size, flags } = r.read_obj().map_err(Error::DecodeMessage)?;
+            (
+                SetxattrIn {
+                    size,
+                    flags,
+                    setxattr_flags: 0,
+                    padding: 0,
+                },
+                size_of::<SetxattrInCompat>(),
+            )
+        };
 
         // The name and value and encoded one after another and separated by a '\0' character.
         let len = (in_header.len as usize)
             .checked_sub(size_of::<InHeader>())
-            .and_then(|l| l.checked_sub(size_of::<SetxattrIn>()))
+            .and_then(|l| l.checked_sub(setxattrin_size))
             .ok_or(Error::InvalidHeaderLength)?;
         let mut buf = vec![0; len];
 
@@ -764,6 +804,7 @@ impl<F: FileSystem + Sync> Server<F> {
             bytes_to_cstr(name)?,
             value,
             flags,
+            SetxattrFlags::from_bits_truncate(setxattr_flags),
         ) {
             Ok(()) => reply_ok(None::<u8>, None, in_header.unique, w),
             Err(e) => reply_error(e, in_header.unique, w),
@@ -871,13 +912,21 @@ impl<F: FileSystem + Sync> Server<F> {
     }
 
     fn init(&self, in_header: InHeader, mut r: Reader, w: Writer) -> Result<usize> {
-        let InitIn {
+        let InitInCompat {
             major,
             minor,
             max_readahead,
             flags,
         } = r.read_obj().map_err(Error::DecodeMessage)?;
 
+        let options = FsOptions::from_bits_truncate(flags as u64);
+
+        let InitInExt { flags2, .. } = if options.contains(FsOptions::INIT_EXT) {
+            r.read_obj().map_err(Error::DecodeMessage)?
+        } else {
+            InitInExt::default()
+        };
+
         if major < KERNEL_VERSION {
             error!("Unsupported fuse protocol version: {}.{}", major, minor);
             return reply_error(
@@ -898,7 +947,7 @@ impl<F: FileSystem + Sync> Server<F> {
             return reply_ok(Some(out), None, in_header.unique, w);
         }
 
-        if minor < KERNEL_MINOR_VERSION {
+        if minor < MIN_KERNEL_MINOR_VERSION {
             error!(
                 "Unsupported fuse protocol minor version: {}.{}",
                 major, minor
@@ -919,28 +968,31 @@ impl<F: FileSystem + Sync> Server<F> {
             | FsOptions::HAS_IOCTL_DIR
             | FsOptions::ATOMIC_O_TRUNC
             | FsOptions::MAX_PAGES
-            | FsOptions::SUBMOUNTS
-            | FsOptions::HANDLE_KILLPRIV_V2;
+            | FsOptions::SUBMOUNTS;
 
-        let capable = FsOptions::from_bits_truncate(flags);
+        let flags_64 = ((flags2 as u64) << 32) | (flags as u64);
+        let capable = FsOptions::from_bits_truncate(flags_64);
 
         let page_size: u32 = unsafe { libc::sysconf(libc::_SC_PAGESIZE).try_into().unwrap() };
         let max_pages = ((MAX_BUFFER_SIZE - 1) / page_size) + 1;
 
         match self.fs.init(capable) {
             Ok(want) => {
-                let enabled = capable & (want | supported);
+                let enabled = (capable & (want | supported)).bits();
+                self.options.store(enabled, Ordering::Relaxed);
 
                 let out = InitOut {
                     major: KERNEL_VERSION,
                     minor: KERNEL_MINOR_VERSION,
                     max_readahead,
-                    flags: enabled.bits(),
-                    max_background: ::std::u16::MAX,
-                    congestion_threshold: (::std::u16::MAX / 4) * 3,
+                    flags: enabled as u32,
+                    max_background: u16::MAX,
+                    congestion_threshold: (u16::MAX / 4) * 3,
                     max_write: MAX_BUFFER_SIZE,
                     time_gran: 1, // nanoseconds
                     max_pages: max_pages.try_into().unwrap(),
+                    map_alignment: 0,
+                    flags2: (enabled >> 32) as u32,
                     ..Default::default()
                 };
 
@@ -1046,7 +1098,7 @@ impl<F: FileSystem + Sync> Server<F> {
             // this only contains integer fields and any value is valid.
             let mut attr = unsafe { MaybeUninit::<libc::stat64>::zeroed().assume_init() };
             attr.st_ino = dir_entry.ino;
-            attr.st_mode = dir_entry.type_;
+            attr.st_mode = dir_entry.type_ << 12;
 
             // We use 0 for the inode value to indicate a negative entry.
             Entry {
@@ -1293,6 +1345,8 @@ impl<F: FileSystem + Sync> Server<F> {
     fn destroy(&self) -> usize {
         // No reply to this function.
         self.fs.destroy();
+        self.options
+            .store(FsOptions::empty().bits(), Ordering::Relaxed);
 
         0
     }
@@ -1432,6 +1486,16 @@ impl<F: FileSystem + Sync> Server<F> {
             Err(e) => reply_error(e, in_header.unique, w),
         }
     }
+
+    fn syncfs(&self, in_header: InHeader, w: Writer) -> Result<usize> {
+        match self
+            .fs
+            .syncfs(Context::from(in_header), in_header.nodeid.into())
+        {
+            Ok(()) => reply_ok(None::<u8>, None, in_header.unique, w),
+            Err(e) => reply_error(e, in_header.unique, w),
+        }
+    }
 }
 
 fn reply_readdir(len: usize, unique: u64, mut w: Writer) -> Result<usize> {
@@ -1515,7 +1579,7 @@ fn add_dirent(
 ) -> io::Result<usize> {
     // Strip the trailing '\0'.
     let name = d.name.to_bytes();
-    if name.len() > ::std::u32::MAX as usize {
+    if name.len() > u32::MAX as usize {
         return Err(io::Error::from_raw_os_error(libc::EOVERFLOW));
     }
 
